@@ -23,7 +23,7 @@
 #ifndef ASYNCSCHEDULE_H
 #define ASYNCSCHEDULE_H
 
-#include "synthschedule.h"
+#include "gslschedule.h"
 #include "datapacket.h"
 #include "weakreference.h"
 
