@@ -20,17 +20,21 @@
 
 #include "gslcommon.h"
 #include "gsldatacache.h"
+#include "gslfilehash.h"
 
-#include <sys/types.h>
-#include <sys/stat.h>
-#include <unistd.h>
 #include <string.h>
 #include <errno.h>
-#include <time.h>
-#include <fcntl.h>
 
 
 /* --- typedefs --- */
+typedef struct {
+  GslDataHandle     dhandle;
+  guint             n_channels;
+  guint             bit_depth;
+  GslLong           n_values;
+  const gfloat     *values;
+  void            (*free_values) (gpointer);
+} MemHandle;
 typedef struct {
   GslDataHandle     dhandle;
   GslDataHandle	   *src_handle;
@@ -41,18 +45,23 @@ typedef struct {
   GslDataHandle	   *src_handle;	/* mirror ChainHandle */
   GslLong	    cut_offset;
   GslLong	    n_cut_values;
+  GslLong	    tail_cut;
 } CutHandle;
 typedef struct {
   GslDataHandle     dhandle;
   GslDataHandle	   *src_handle;	/* mirror ChainHandle */
+  GslLong	    requested_paste_offset;
   GslLong	    paste_offset;
   GslLong	    n_paste_values;
+  guint		    paste_bit_depth;
   const gfloat	   *paste_values;
   void            (*free_values) (gpointer);
 } InsertHandle;
 typedef struct {
   GslDataHandle     dhandle;
   GslDataHandle	   *src_handle;	/* mirror ChainHandle */
+  GslLong	    requested_first;
+  GslLong	    requested_last;
   GslLong	    loop_start;
   GslLong	    loop_width;
 } LoopHandle;
@@ -63,14 +72,34 @@ typedef struct {
 } DCacheHandle;
 typedef struct {
   GslDataHandle     dhandle;
+  guint		    n_channels;
   GslWaveFormatType format;
   guint             byte_order;
-  guint             boffset_reminder;
-  gint              fd;
+  GslLong	    byte_offset;
+  GslLong	    requested_length;
+  GslHFile	   *hfile;
 } WaveHandle;
 
 
 /* --- standard functions --- */
+gboolean
+gsl_data_handle_common_init (GslDataHandle *dhandle,
+			     const gchar   *file_name)
+{
+  g_return_val_if_fail (dhandle != NULL, FALSE);
+  g_return_val_if_fail (dhandle->vtable == NULL, FALSE);
+  g_return_val_if_fail (dhandle->name == NULL, FALSE);
+  g_return_val_if_fail (dhandle->ref_count == 0, FALSE);
+  
+  dhandle->name = g_strdup (file_name);
+  gsl_mutex_init (&dhandle->mutex);
+  dhandle->ref_count = 1;
+  dhandle->open_count = 0;
+  memset (&dhandle->setup, 0, sizeof (dhandle->setup));
+  
+  return TRUE;
+}
+
 GslDataHandle*
 gsl_data_handle_ref (GslDataHandle *dhandle)
 {
@@ -109,22 +138,37 @@ gsl_data_handle_unref (GslDataHandle *dhandle)
   destroy = dhandle->ref_count == 0;
   GSL_SPIN_UNLOCK (&dhandle->mutex);
   if (destroy)
-    dhandle->vtable->destroy (dhandle);
+    {
+      g_return_if_fail (dhandle->open_count == 0);
+      dhandle->vtable->destroy (dhandle);
+    }
 }
 
-gint	/* errno return */
+GslErrorType
 gsl_data_handle_open (GslDataHandle *dhandle)
 {
-  g_return_val_if_fail (dhandle != NULL, EINVAL);
-  g_return_val_if_fail (dhandle->ref_count > 0, EINVAL);
+  g_return_val_if_fail (dhandle != NULL, GSL_ERROR_INTERNAL);
+  g_return_val_if_fail (dhandle->ref_count > 0, GSL_ERROR_INTERNAL);
 
   GSL_SPIN_LOCK (&dhandle->mutex);
   if (dhandle->open_count == 0)
     {
-      gint error = dhandle->vtable->open (dhandle);
-      
+      GslErrorType error;
+
+      memset (&dhandle->setup, 0, sizeof (dhandle->setup));
+      error = dhandle->vtable->open (dhandle, &dhandle->setup);
+      if (!error && (dhandle->setup.n_values < 0 ||
+		     dhandle->setup.n_channels < 1 ||
+		     dhandle->setup.bit_depth < 1))
+	{
+	  g_warning ("internal error in data handle open() (%p): nv=%ld nc=%u bd=%u",
+		     dhandle->vtable->open, dhandle->setup.n_values, dhandle->setup.n_channels, dhandle->setup.bit_depth);
+	  dhandle->vtable->close (dhandle);
+	  error = GSL_ERROR_INTERNAL;
+	}
       if (error)
 	{
+	  memset (&dhandle->setup, 0, sizeof (dhandle->setup));
 	  GSL_SPIN_UNLOCK (&dhandle->mutex);
 	  return error;
 	}
@@ -135,7 +179,7 @@ gsl_data_handle_open (GslDataHandle *dhandle)
     dhandle->open_count++;
   GSL_SPIN_UNLOCK (&dhandle->mutex);
   
-  return 0;
+  return GSL_ERROR_NONE;
 }
 
 void
@@ -166,15 +210,14 @@ gsl_data_handle_read (GslDataHandle *dhandle,
   GslLong l;
   
   g_return_val_if_fail (dhandle != NULL, -1);
-  g_return_val_if_fail (dhandle->ref_count > 0, -1);
   g_return_val_if_fail (dhandle->open_count > 0, -1);
   g_return_val_if_fail (value_offset >= 0, -1);
   if (n_values < 1)
     return 0;
   g_return_val_if_fail (values != NULL, -1);
-  g_return_val_if_fail (value_offset < dhandle->n_values, -1);
+  g_return_val_if_fail (value_offset < dhandle->setup.n_values, -1);
   
-  n_values = MIN (n_values, dhandle->n_values - value_offset);
+  n_values = MIN (n_values, dhandle->setup.n_values - value_offset);
   GSL_SPIN_LOCK (&dhandle->mutex);
   l = dhandle->vtable->read (dhandle, value_offset, n_values, values);
   GSL_SPIN_UNLOCK (&dhandle->mutex);
@@ -182,68 +225,183 @@ gsl_data_handle_read (GslDataHandle *dhandle,
   return l;
 }
 
-gboolean
-gsl_data_handle_common_init (GslDataHandle *dhandle,
-			     const gchar   *file_name,
-			     guint          bit_depth)
+GslLong
+gsl_data_handle_length (GslDataHandle *dhandle)
 {
-  g_return_val_if_fail (dhandle != NULL, FALSE);
-  g_return_val_if_fail (dhandle->vtable == NULL, FALSE);
-  g_return_val_if_fail (dhandle->name == NULL, FALSE);
-  g_return_val_if_fail (dhandle->ref_count == 0, FALSE);
-  g_return_val_if_fail (bit_depth > 0, FALSE);
+  GslLong l;
+
+  g_return_val_if_fail (dhandle != NULL, 0);
+  g_return_val_if_fail (dhandle->open_count > 0, 0);
+
+  GSL_SPIN_LOCK (&dhandle->mutex);
+  l = dhandle->open_count ? dhandle->setup.n_values : 0;
+  GSL_SPIN_UNLOCK (&dhandle->mutex);
+
+  return l;
+}
+
+guint
+gsl_data_handle_n_channels (GslDataHandle *dhandle)
+{
+  guint n;
+
+  g_return_val_if_fail (dhandle != NULL, 0);
+  g_return_val_if_fail (dhandle->open_count > 0, 0);
+
+  GSL_SPIN_LOCK (&dhandle->mutex);
+  n = dhandle->open_count ? dhandle->setup.n_channels : 0;
+  GSL_SPIN_UNLOCK (&dhandle->mutex);
+
+  return n;
+}
+
+guint
+gsl_data_handle_bit_depth (GslDataHandle *dhandle)
+{
+  guint n;
+
+  g_return_val_if_fail (dhandle != NULL, 0);
+  g_return_val_if_fail (dhandle->open_count > 0, 0);
+
+  GSL_SPIN_LOCK (&dhandle->mutex);
+  n = dhandle->open_count ? dhandle->setup.bit_depth : 0;
+  GSL_SPIN_UNLOCK (&dhandle->mutex);
+
+  return n;
+}
+
+const gchar*
+gsl_data_handle_name (GslDataHandle *dhandle)
+{
+  g_return_val_if_fail (dhandle != NULL, NULL);
+
+  return dhandle->name;
+}
+
+
+/* --- const memory handle --- */
+static GslErrorType
+mem_handle_open (GslDataHandle      *dhandle,
+		 GslDataHandleSetup *setup)
+{
+  MemHandle *mhandle = (MemHandle*) dhandle;
+
+  setup->n_values = mhandle->n_values;
+  setup->n_channels = mhandle->n_channels;
+  setup->bit_depth = mhandle->bit_depth;
+
+  return GSL_ERROR_NONE;
+}
+
+static void
+mem_handle_close (GslDataHandle *dhandle)
+{
+  /* MemHandle *mhandle = (MemHandle*) dhandle; */
+}
+
+static void
+mem_handle_destroy (GslDataHandle *dhandle)
+{
+  MemHandle *mhandle = (MemHandle*) dhandle;
+  void (*free_values) (gpointer) = mhandle->free_values;
+  const gfloat *mem_values = mhandle->values;
+
+  gsl_data_handle_common_free (dhandle);
+  mhandle->values = NULL;
+  mhandle->free_values = NULL;
+  gsl_delete_struct (MemHandle, mhandle);
   
-  if (file_name)
+  if (free_values)
+    free_values ((gpointer) mem_values);
+}
+
+static GslLong
+mem_handle_read (GslDataHandle *dhandle,
+		 GslLong        voffset,
+		 GslLong        n_values,
+		 gfloat        *values)
+{
+  MemHandle *mhandle = (MemHandle*) dhandle;
+
+  g_return_val_if_fail (voffset + n_values <= mhandle->n_values, -1);
+
+  memcpy (values, mhandle->values + voffset, n_values * sizeof (values[0]));
+
+  return n_values;
+}
+
+GslDataHandle*
+gsl_data_handle_new_mem (guint         n_channels,
+			 guint         bit_depth,
+			 GslLong       n_values,
+			 const gfloat *values,
+			 void        (*free) (gpointer values))
+{
+  static GslDataHandleFuncs mem_handle_vtable = {
+    mem_handle_open,
+    mem_handle_read,
+    mem_handle_close,
+    mem_handle_destroy,
+  };
+  MemHandle *mhandle;
+  gboolean success;
+
+  g_return_val_if_fail (n_channels > 0, NULL);
+  g_return_val_if_fail (bit_depth > 0, NULL);
+  g_return_val_if_fail (n_values >= n_channels, NULL);
+  if (n_values)
+    g_return_val_if_fail (values != NULL, NULL);
+  
+  mhandle = gsl_new_struct0 (MemHandle, 1);
+  success = gsl_data_handle_common_init (&mhandle->dhandle, NULL);
+  if (success)
     {
-      struct stat statbuf = { 0, };
-      
-      if (stat (file_name, &statbuf) < 0 || statbuf.st_size < 1)
-	return FALSE;
-      
-      dhandle->name = g_strdup (file_name);
-      dhandle->mtime = statbuf.st_mtime;
-      dhandle->n_values = statbuf.st_size;
+      mhandle->dhandle.name = g_strconcat ("// #memory /", NULL);
+      mhandle->dhandle.vtable = &mem_handle_vtable;
+      mhandle->n_channels = n_channels;
+      mhandle->bit_depth = bit_depth;
+      mhandle->n_values = n_values / mhandle->n_channels;
+      mhandle->n_values *= mhandle->n_channels;
+      mhandle->values = values;
+      mhandle->free_values = free;
     }
   else
     {
-      dhandle->name = NULL;
-      dhandle->mtime = time (NULL);
-      dhandle->n_values = 0;
+      gsl_delete_struct (MemHandle, mhandle);
+      return NULL;
     }
-  dhandle->bit_depth = bit_depth;
-  gsl_mutex_init (&dhandle->mutex);
-  dhandle->ref_count = 1;
-  dhandle->open_count = 0;
-  
-  return TRUE;
+  return &mhandle->dhandle;
 }
 
 
 /* --- chain handle --- */
-static gint
-chain_handle_open (GslDataHandle *data_handle)
+static GslErrorType
+chain_handle_open (GslDataHandle      *dhandle,
+		   GslDataHandleSetup *setup)
 {
-  ChainHandle *chandle = (ChainHandle*) data_handle;
+  ChainHandle *chandle = (ChainHandle*) dhandle;
+  GslErrorType error;
 
-  if (chandle->src_handle)
-    return gsl_data_handle_open (chandle->src_handle);
-  else
-    return 0;
+  error = gsl_data_handle_open (chandle->src_handle);
+  if (error != GSL_ERROR_NONE)
+    return error;
+  *setup = chandle->src_handle->setup;
+
+  return GSL_ERROR_NONE;
 }
 
 static void
-chain_handle_close (GslDataHandle *data_handle)
+chain_handle_close (GslDataHandle *dhandle)
 {
-  ChainHandle *chandle = (ChainHandle*) data_handle;
+  ChainHandle *chandle = (ChainHandle*) dhandle;
   
-  if (chandle->src_handle)
-    gsl_data_handle_close (chandle->src_handle);
+  gsl_data_handle_close (chandle->src_handle);
 }
 
 
 /* --- reversed handle --- */
 static void
-reversed_handle_destroy (GslDataHandle *data_handle)
+reverse_handle_destroy (GslDataHandle *data_handle)
 {
   ReversedHandle *rhandle = (ReversedHandle*) data_handle;
   
@@ -254,13 +412,13 @@ reversed_handle_destroy (GslDataHandle *data_handle)
 }
 
 static GslLong
-reversed_handle_read (GslDataHandle *data_handle,
-		      GslLong        voffset,
-		      GslLong        n_values,
-		      gfloat        *values)
+reverse_handle_read (GslDataHandle *dhandle,
+		     GslLong        voffset,
+		     GslLong        n_values,
+		     gfloat        *values)
 {
-  ReversedHandle *rhandle = (ReversedHandle*) data_handle;
-  GslLong left, new_offset = data_handle->n_values - (voffset + n_values);
+  ReversedHandle *rhandle = (ReversedHandle*) dhandle;
+  GslLong left, new_offset = dhandle->setup.n_values - (voffset + n_values);
   gfloat *t, *p = values;
   
   g_assert (new_offset >= 0);
@@ -292,13 +450,13 @@ reversed_handle_read (GslDataHandle *data_handle,
 }
 
 GslDataHandle*
-gsl_data_handle_new_reversed (GslDataHandle *src_handle)
+gsl_data_handle_new_reverse (GslDataHandle *src_handle)
 {
-  static GslDataHandleFuncs reversed_handle_vtable = {
+  static GslDataHandleFuncs reverse_handle_vtable = {
     chain_handle_open,
-    reversed_handle_read,
+    reverse_handle_read,
     chain_handle_close,
-    reversed_handle_destroy,
+    reverse_handle_destroy,
   };
   ReversedHandle *rhandle;
   gboolean success;
@@ -306,12 +464,11 @@ gsl_data_handle_new_reversed (GslDataHandle *src_handle)
   g_return_val_if_fail (src_handle != NULL, NULL);
   
   rhandle = gsl_new_struct0 (ReversedHandle, 1);
-  success = gsl_data_handle_common_init (&rhandle->dhandle, NULL, src_handle->bit_depth);
+  success = gsl_data_handle_common_init (&rhandle->dhandle, NULL);
   if (success)
     {
       rhandle->dhandle.name = g_strconcat (src_handle->name, "// #reversed /", NULL);
-      rhandle->dhandle.vtable = &reversed_handle_vtable;
-      rhandle->dhandle.n_values = src_handle->n_values;
+      rhandle->dhandle.vtable = &reverse_handle_vtable;
       rhandle->src_handle = gsl_data_handle_ref (src_handle);
     }
   else
@@ -324,6 +481,23 @@ gsl_data_handle_new_reversed (GslDataHandle *src_handle)
 
 
 /* --- cut handle --- */
+static GslErrorType
+cut_handle_open (GslDataHandle      *dhandle,
+		 GslDataHandleSetup *setup)
+{
+  CutHandle *chandle = (CutHandle*) dhandle;
+  GslErrorType error;
+
+  error = gsl_data_handle_open (chandle->src_handle);
+  if (error != GSL_ERROR_NONE)
+    return error;
+  *setup = chandle->src_handle->setup;
+  setup->n_values -= MIN (setup->n_values, chandle->tail_cut);
+  setup->n_values -= MIN (setup->n_values, chandle->n_cut_values);
+
+  return GSL_ERROR_NONE;
+}
+
 static void
 cut_handle_destroy (GslDataHandle *data_handle)
 {
@@ -336,12 +510,12 @@ cut_handle_destroy (GslDataHandle *data_handle)
 }
 
 static GslLong
-cut_handle_read (GslDataHandle *data_handle,
+cut_handle_read (GslDataHandle *dhandle,
 		 GslLong        voffset,
 		 GslLong        n_values,
 		 gfloat        *values)
 {
-  CutHandle *chandle = (CutHandle*) data_handle;
+  CutHandle *chandle = (CutHandle*) dhandle;
   GslLong orig_n_values = n_values;
   
   if (voffset < chandle->cut_offset)
@@ -356,7 +530,7 @@ cut_handle_read (GslDataHandle *data_handle,
       voffset += l;
     }
   
-  if (voffset >= chandle->cut_offset)
+  if (voffset >= chandle->cut_offset && n_values)
     {
       GslLong l = gsl_data_handle_read (chandle->src_handle, voffset + chandle->n_cut_values, n_values, values);
       
@@ -371,14 +545,14 @@ cut_handle_read (GslDataHandle *data_handle,
   return orig_n_values - n_values;
 }
 
-GslDataHandle*
+static GslDataHandle*
 gsl_data_handle_new_translate (GslDataHandle *src_handle,
 			       GslLong        cut_offset,
 			       GslLong        n_cut_values,
 			       GslLong	      tail_cut)
 {
   static GslDataHandleFuncs cut_handle_vtable = {
-    chain_handle_open,
+    cut_handle_open,
     cut_handle_read,
     chain_handle_close,
     cut_handle_destroy,
@@ -388,19 +562,17 @@ gsl_data_handle_new_translate (GslDataHandle *src_handle,
   
   g_return_val_if_fail (src_handle != NULL, NULL);
   g_return_val_if_fail (cut_offset >= 0 && n_cut_values >= 0 && tail_cut >= 0, NULL);
-  g_return_val_if_fail (cut_offset < src_handle->n_values, NULL);
-  g_return_val_if_fail (cut_offset + n_cut_values + tail_cut < src_handle->n_values, NULL);
   
   chandle = gsl_new_struct0 (CutHandle, 1);
-  success = gsl_data_handle_common_init (&chandle->dhandle, NULL, src_handle->bit_depth);
+  success = gsl_data_handle_common_init (&chandle->dhandle, NULL);
   if (success)
     {
       chandle->dhandle.name = g_strconcat (src_handle->name, "// #translate /", NULL);
       chandle->dhandle.vtable = &cut_handle_vtable;
-      chandle->dhandle.n_values = src_handle->n_values - n_cut_values - tail_cut;
       chandle->src_handle = gsl_data_handle_ref (src_handle);
       chandle->cut_offset = n_cut_values ? cut_offset : 0;
       chandle->n_cut_values = n_cut_values;
+      chandle->tail_cut = tail_cut;
     }
   else
     {
@@ -410,8 +582,66 @@ gsl_data_handle_new_translate (GslDataHandle *src_handle,
   return &chandle->dhandle;
 }
 
+/**
+ * gsl_data_handle_new_cut
+ * @src_handle:   source GslDataHandle
+ * @cut_offset:   offset of gap into @src_handle
+ * @n_cut_values: length of gap in @src_handle
+ * @RETURNS: a newly created data handle
+ *
+ * Create a new data handle containing the contents of @src_handle
+ * minus @n_cut_values at offset @cut_offset.
+ */
+GslDataHandle*
+gsl_data_handle_new_cut (GslDataHandle *src_handle,
+			 GslLong        cut_offset,
+			 GslLong        n_cut_values)
+{
+  return gsl_data_handle_new_translate (src_handle, cut_offset, n_cut_values, 0);
+}
+
+/**
+ * gsl_data_handle_new_crop
+ * @src_handle: source GslDataHandle
+ * @n_head_cut: number of values to cut at data handle head
+ * @n_tail_cut: number of values to cut at data handle tail
+ * @RETURNS: a newly created data handle
+ *
+ * Create a new data handle containing the contents of @src_handle
+ * minus @n_head_cut values at the start and @n_tail_cut values at
+ * the end.
+ */
+GslDataHandle*
+gsl_data_handle_new_crop (GslDataHandle *src_handle,
+			  GslLong        n_head_cut,
+			  GslLong        n_tail_cut)
+{
+  return gsl_data_handle_new_translate (src_handle, 0, n_head_cut, n_tail_cut);
+}
+
 
 /* --- insert handle --- */
+static GslErrorType
+insert_handle_open (GslDataHandle      *dhandle,
+		    GslDataHandleSetup *setup)
+{
+  InsertHandle *ihandle = (InsertHandle*) dhandle;
+  GslErrorType error;
+
+  error = gsl_data_handle_open (ihandle->src_handle);
+  if (error != GSL_ERROR_NONE)
+    return error;
+  *setup = ihandle->src_handle->setup;
+  ihandle->paste_offset = ihandle->requested_paste_offset < 0 ? setup->n_values : ihandle->requested_paste_offset;
+  if (setup->n_values < ihandle->paste_offset)
+    setup->n_values = ihandle->paste_offset + ihandle->n_paste_values;
+  else
+    setup->n_values += ihandle->n_paste_values;
+  setup->bit_depth = MAX (setup->bit_depth, ihandle->paste_bit_depth);
+
+  return GSL_ERROR_NONE;
+}
+
 static void
 insert_handle_destroy (GslDataHandle *data_handle)
 {
@@ -419,8 +649,7 @@ insert_handle_destroy (GslDataHandle *data_handle)
   void (*free_values) (gpointer) = ihandle->free_values;
   const gfloat *paste_values = ihandle->paste_values;
   
-  if (ihandle->src_handle)
-    gsl_data_handle_unref (ihandle->src_handle);
+  gsl_data_handle_unref (ihandle->src_handle);
   
   gsl_data_handle_common_free (data_handle);
   ihandle->paste_values = NULL;
@@ -440,9 +669,10 @@ insert_handle_read (GslDataHandle *data_handle,
   InsertHandle *ihandle = (InsertHandle*) data_handle;
   GslLong l, orig_n_values = n_values;
   
-  if (voffset < ihandle->paste_offset)
+  if (voffset < ihandle->src_handle->setup.n_values &&
+      voffset < ihandle->paste_offset)
     {
-      l = MIN (n_values, ihandle->paste_offset - voffset);
+      l = MIN (n_values, MIN (ihandle->paste_offset, ihandle->src_handle->setup.n_values) - voffset);
       l = gsl_data_handle_read (ihandle->src_handle, voffset, l, values);
       if (l < 0)
 	return l;       /* pass on errors */
@@ -451,8 +681,17 @@ insert_handle_read (GslDataHandle *data_handle,
       n_values -= l;
       values += l;
     }
-  
-  if (voffset >= ihandle->paste_offset && voffset < ihandle->paste_offset + ihandle->n_paste_values)
+
+  if (n_values && voffset >= ihandle->src_handle->setup.n_values && voffset < ihandle->paste_offset)
+    {
+      l = MIN (n_values, ihandle->paste_offset - voffset);
+      memset (values, 0, l * sizeof (values[0]));
+      voffset += l;
+      n_values -= l;
+      values += l;
+    }
+
+  if (n_values && voffset >= ihandle->paste_offset && voffset < ihandle->paste_offset + ihandle->n_paste_values)
     {
       l = MIN (n_values, ihandle->paste_offset + ihandle->n_paste_values - voffset);
       memcpy (values, ihandle->paste_values + voffset - ihandle->paste_offset, l * sizeof (values[0]));
@@ -461,14 +700,13 @@ insert_handle_read (GslDataHandle *data_handle,
       values += l;
     }
   
-  if (voffset >= ihandle->paste_offset + ihandle->n_paste_values && n_values)
+  if (n_values && voffset >= ihandle->paste_offset + ihandle->n_paste_values)
     {
       l = gsl_data_handle_read (ihandle->src_handle, voffset - ihandle->n_paste_values, n_values, values);
       if (l < 0 && orig_n_values == n_values)
 	return l;       /* pass on errors */
       else if (l < 0)
 	l = 0;
-      
       n_values -= l;
     }
   
@@ -484,7 +722,7 @@ gsl_data_handle_new_insert (GslDataHandle *src_handle,
 			    void         (*free) (gpointer values))
 {
   static GslDataHandleFuncs insert_handle_vtable = {
-    chain_handle_open,
+    insert_handle_open,
     insert_handle_read,
     chain_handle_close,
     insert_handle_destroy,
@@ -492,33 +730,22 @@ gsl_data_handle_new_insert (GslDataHandle *src_handle,
   InsertHandle *ihandle;
   gboolean success;
   
+  g_return_val_if_fail (src_handle != NULL, NULL);
   g_return_val_if_fail (n_paste_values >= 0, NULL);
-  if (!src_handle)
-    {
-      g_return_val_if_fail (insertion_offset == 0, NULL);
-      g_return_val_if_fail (n_paste_values > 0, NULL);
-    }
-  else
-    {
-      if (insertion_offset < 0)
-	insertion_offset = src_handle->n_values;
-      g_return_val_if_fail (insertion_offset <= src_handle->n_values, NULL);
-    }
   if (n_paste_values)
     g_return_val_if_fail (paste_values != NULL, NULL);
   
   ihandle = gsl_new_struct0 (InsertHandle, 1);
-  if (src_handle)
-    paste_bit_depth = MAX (src_handle->bit_depth, paste_bit_depth);
-  success = gsl_data_handle_common_init (&ihandle->dhandle, NULL, paste_bit_depth);
+  success = gsl_data_handle_common_init (&ihandle->dhandle, NULL);
   if (success)
     {
       ihandle->dhandle.name = g_strconcat (src_handle ? src_handle->name : "", "// #insert /", NULL);
       ihandle->dhandle.vtable = &insert_handle_vtable;
-      ihandle->dhandle.n_values = n_paste_values + (src_handle ? src_handle->n_values : 0);
-      ihandle->src_handle = src_handle ? gsl_data_handle_ref (src_handle) : NULL;
-      ihandle->paste_offset = insertion_offset;
+      ihandle->src_handle = gsl_data_handle_ref (src_handle);
+      ihandle->requested_paste_offset = insertion_offset;
+      ihandle->paste_offset = 0;
       ihandle->n_paste_values = n_paste_values;
+      ihandle->paste_bit_depth = paste_bit_depth;
       ihandle->paste_values = paste_values;
       ihandle->free_values = free;
     }
@@ -532,6 +759,33 @@ gsl_data_handle_new_insert (GslDataHandle *src_handle,
 
 
 /* --- loop handle --- */
+static GslErrorType
+loop_handle_open (GslDataHandle      *dhandle,
+		  GslDataHandleSetup *setup)
+{
+  LoopHandle *lhandle = (LoopHandle*) dhandle;
+  GslErrorType error;
+
+  error = gsl_data_handle_open (lhandle->src_handle);
+  if (error != GSL_ERROR_NONE)
+    return error;
+
+  *setup = lhandle->src_handle->setup;
+  if (setup->n_values > lhandle->requested_last)
+    {
+      lhandle->loop_start = lhandle->requested_first;
+      lhandle->loop_width = lhandle->requested_last - lhandle->requested_first + 1;
+      setup->n_values = GSL_MAXLONG;
+    }
+  else	/* cannot loop */
+    {
+      lhandle->loop_start = setup->n_values;
+      lhandle->loop_width = 0;
+    }
+
+  return GSL_ERROR_NONE;
+}
+
 static void
 loop_handle_destroy (GslDataHandle *data_handle)
 {
@@ -570,11 +824,11 @@ loop_handle_read (GslDataHandle *data_handle,
 
 GslDataHandle*
 gsl_data_handle_new_looped (GslDataHandle *src_handle,
-			    GslLong        loop_start,
-			    GslLong        loop_end)
+			    GslLong        loop_first,
+			    GslLong        loop_last)
 {
   static GslDataHandleFuncs loop_handle_vtable = {
-    chain_handle_open,
+    loop_handle_open,
     loop_handle_read,
     chain_handle_close,
     loop_handle_destroy,
@@ -583,20 +837,20 @@ gsl_data_handle_new_looped (GslDataHandle *src_handle,
   gboolean success;
   
   g_return_val_if_fail (src_handle != NULL, NULL);
-  g_return_val_if_fail (loop_start >= 0, NULL);
-  g_return_val_if_fail (loop_start <= loop_end, NULL);
-  g_return_val_if_fail (loop_end < src_handle->n_values, NULL);
+  g_return_val_if_fail (loop_first >= 0, NULL);
+  g_return_val_if_fail (loop_last >= loop_first, NULL);
   
   lhandle = gsl_new_struct0 (LoopHandle, 1);
-  success = gsl_data_handle_common_init (&lhandle->dhandle, NULL, src_handle->bit_depth);
+  success = gsl_data_handle_common_init (&lhandle->dhandle, NULL);
   if (success)
     {
-      lhandle->dhandle.name = g_strdup_printf ("%s// #loop(0x%lx:0x%lx) /", src_handle->name, loop_start, loop_end);
+      lhandle->dhandle.name = g_strdup_printf ("%s// #loop(0x%lx:0x%lx) /", src_handle->name, loop_first, loop_last);
       lhandle->dhandle.vtable = &loop_handle_vtable;
-      lhandle->dhandle.n_values = GSL_MAXLONG;
       lhandle->src_handle = gsl_data_handle_ref (src_handle);
-      lhandle->loop_start = loop_start;
-      lhandle->loop_width = loop_end - loop_start + 1;
+      lhandle->requested_first = loop_first;
+      lhandle->requested_last = loop_last;
+      lhandle->loop_start = 0;
+      lhandle->loop_width = 0;
     }
   else
     {
@@ -619,14 +873,21 @@ dcache_handle_destroy (GslDataHandle *data_handle)
   gsl_delete_struct (DCacheHandle, dhandle);
 }
 
-static gint
-dcache_handle_open (GslDataHandle *data_handle)
+static GslErrorType
+dcache_handle_open (GslDataHandle      *dhandle,
+		    GslDataHandleSetup *setup)
 {
-  DCacheHandle *dhandle = (DCacheHandle*) data_handle;
-  
-  gsl_data_cache_open (dhandle->dcache);
-  
-  return 0;	/* FIXME: should catch messages here and return error */
+  DCacheHandle *chandle = (DCacheHandle*) dhandle;
+  GslErrorType error;
+
+  error = gsl_data_handle_open (chandle->dcache->dhandle);
+  if (error != GSL_ERROR_NONE)
+    return error;
+  gsl_data_cache_open (chandle->dcache);
+  *setup = chandle->dcache->dhandle->setup;
+  gsl_data_handle_close (chandle->dcache->dhandle);
+
+  return GSL_ERROR_NONE;
 }
 
 static void
@@ -669,14 +930,13 @@ gsl_data_handle_new_dcached (GslDataCache *dcache)
   g_return_val_if_fail (dcache != NULL, NULL);
   
   dhandle = gsl_new_struct0 (DCacheHandle, 1);
-  success = gsl_data_handle_common_init (&dhandle->dhandle, NULL, dcache->handle->bit_depth);
+  success = gsl_data_handle_common_init (&dhandle->dhandle, NULL);
   if (success)
     {
-      dhandle->dhandle.name = g_strdup_printf ("%s// #dcache /", dcache->handle->name);
+      dhandle->dhandle.name = g_strdup_printf ("%s// #dcache /", dcache->dhandle->name);
       dhandle->dhandle.vtable = &dcache_handle_vtable;
-      dhandle->dhandle.n_values = dcache->handle->n_values;
       dhandle->dcache = gsl_data_cache_ref (dcache);
-      dhandle->node_size = GSL_DATA_CACHE_NODE_SIZE (dcache);
+      dhandle->node_size = GSL_DATA_CACHE_NODE_SIZE (dcache) + dcache->padding;
     }
   else
     {
@@ -719,26 +979,44 @@ wave_handle_destroy (GslDataHandle *data_handle)
   gsl_delete_struct (WaveHandle, whandle);
 }
 
-static gint
-wave_handle_open (GslDataHandle *data_handle)
+static GslErrorType
+wave_handle_open (GslDataHandle      *data_handle,
+		  GslDataHandleSetup *setup)
 {
   WaveHandle *whandle = (WaveHandle*) data_handle;
-  
-  whandle->fd = open (whandle->dhandle.name, O_RDONLY);
-  
-  if (whandle->fd < 0)
-    return errno ? errno : EIO;
-  
-  return 0;
+
+  whandle->hfile = gsl_hfile_open (whandle->dhandle.name);
+  if (!whandle->hfile)
+    return gsl_error_from_errno (errno, GSL_ERROR_OPEN_FAILED);
+  else
+    {
+      GslLong l, fwidth = wave_format_byte_width (whandle->format);
+      /* convert size into n_values, i.e. float length */
+      l = whandle->hfile->n_bytes;
+      l -= MIN (l, whandle->byte_offset);
+      if (l >= fwidth)
+	{
+	  l /= fwidth;
+	  if (whandle->requested_length < 0)
+	    setup->n_values = l;
+	  else
+	    setup->n_values = MIN (l, whandle->requested_length);
+	}
+      else
+	setup->n_values = 0;
+      setup->n_channels = whandle->n_channels;
+      setup->bit_depth = wave_format_bit_depth (whandle->format);
+      return GSL_ERROR_NONE;
+    }
 }
 
 static void
-wave_handle_close (GslDataHandle *data_handle)
+wave_handle_close (GslDataHandle *dhandle)
 {
-  WaveHandle *whandle = (WaveHandle*) data_handle;
+  WaveHandle *whandle = (WaveHandle*) dhandle;
   
-  close (whandle->fd);
-  whandle->fd = -1;
+  gsl_hfile_close (whandle->hfile);
+  whandle->hfile = NULL;
 }
 
 static GslLong
@@ -749,22 +1027,17 @@ wave_handle_read (GslDataHandle *data_handle,
 {
   WaveHandle *whandle = (WaveHandle*) data_handle;
   gpointer buffer = values;
-  GslLong l, i = 0;
-  
-  i = voffset * wave_format_byte_width (whandle->format);	/* float offset into bytes */
-  i += whandle->boffset_reminder;				/* remaining byte offset */
-  l = lseek (whandle->fd, i, SEEK_SET);
-  if (l < 0 && errno != EINVAL)
-    return -1;
-  if (l != i)
-    return 0;
-  
+  GslLong l, i, byte_offset;
+
+  byte_offset = voffset * wave_format_byte_width (whandle->format);	/* float offset into bytes */
+  byte_offset += whandle->byte_offset;
+
   switch (whandle->format)
     {
       guint8 *u8; gint8 *s8; guint16 *u16; guint32 *u32;
     case GSL_WAVE_FORMAT_UNSIGNED_8:
       u8 = buffer; u8 += n_values * 3;
-      l = read (whandle->fd, u8, n_values);
+      l = gsl_hfile_pread (whandle->hfile, byte_offset, n_values, u8);
       if (l < 1)
 	return l;
       for (i = 0; i < l; i++)
@@ -775,7 +1048,7 @@ wave_handle_read (GslDataHandle *data_handle,
       break;
     case GSL_WAVE_FORMAT_SIGNED_8:
       s8 = buffer; s8 += n_values * 3;
-      l = read (whandle->fd, s8, n_values);
+      l = gsl_hfile_pread (whandle->hfile, byte_offset, n_values, s8);
       if (l < 1)
 	return l;
       for (i = 0; i < l; i++)
@@ -786,7 +1059,7 @@ wave_handle_read (GslDataHandle *data_handle,
     case GSL_WAVE_FORMAT_SIGNED_16:
     case GSL_WAVE_FORMAT_UNSIGNED_16:
       u16 = buffer; u16 += n_values;
-      l = read (whandle->fd, u16, n_values << 1);
+      l = gsl_hfile_pread (whandle->hfile, byte_offset, n_values << 1, u16);
       if (l < 2)
 	return l < 0 ? l : 0;
       l >>= 1;
@@ -854,7 +1127,7 @@ wave_handle_read (GslDataHandle *data_handle,
       break;
     case GSL_WAVE_FORMAT_FLOAT:
       u32 = buffer;
-      l = read (whandle->fd, u32, n_values << 2);
+      l = gsl_hfile_pread (whandle->hfile, byte_offset, n_values << 2, u32);
       if (l < 4)
 	return l < 0 ? l : 0;
       l >>= 2;
@@ -863,17 +1136,20 @@ wave_handle_read (GslDataHandle *data_handle,
 	  u32[i] = GUINT32_SWAP_LE_BE (u32[i]);
       break;
     default:
+      l = -1;
       g_assert_not_reached ();
     }
   
   return l;
 }
 
-static GslDataHandle*
-raw_wave_handle_new_cached (const gchar      *file_name,
-			    GslWaveFormatType format,
-			    guint	      byte_order,
-			    guint	      boffset_reminder)
+GslDataHandle*
+gsl_wave_handle_new (const gchar      *file_name,
+		     guint             n_channels,
+		     GslWaveFormatType format,
+		     guint             byte_order,
+		     GslLong           byte_offset,
+		     GslLong           n_values)
 {
   static GslDataHandleFuncs wave_handle_vtable = {
     wave_handle_open,
@@ -881,111 +1157,32 @@ raw_wave_handle_new_cached (const gchar      *file_name,
     wave_handle_close,
     wave_handle_destroy,
   };
-  GslDataHandleHash hash = { 0, };
-  GslDataHandle *cache_handle;
   WaveHandle *whandle;
-  gboolean success;
-  
-  g_return_val_if_fail (file_name != NULL, NULL);
-  g_return_val_if_fail (format > GSL_WAVE_FORMAT_NONE && format < GSL_WAVE_FORMAT_LAST, NULL);
-  g_return_val_if_fail (byte_order == G_LITTLE_ENDIAN || byte_order == G_BIG_ENDIAN, NULL);
-  g_return_val_if_fail (boffset_reminder < wave_format_byte_width (format), NULL);
 
-  hash.fpointer = &wave_handle_vtable;
-  hash.strings[0] = (gchar*) file_name;
-  hash.data[0].v_long = format;
-  hash.data[1].v_long = byte_order;
-  hash.data[2].v_long = boffset_reminder;
-  cache_handle = gsl_data_handle_cached (hash);
-  if (cache_handle)
-    return cache_handle;
-
-  whandle = gsl_new_struct0 (WaveHandle, 1);
-  success = gsl_data_handle_common_init (&whandle->dhandle, file_name, wave_format_bit_depth (format));
-  if (success)
-    {
-      GslLong l;
-      
-      whandle->dhandle.vtable = &wave_handle_vtable;
-      whandle->format = format;
-      whandle->byte_order = byte_order;
-      whandle->boffset_reminder = boffset_reminder;
-      whandle->fd = -1;
-      
-      /* convert size into n_values, i.e. float length */
-      l = whandle->dhandle.n_values;
-      if (l > whandle->boffset_reminder)
-	{
-	  l -= whandle->boffset_reminder;
-	  l /= wave_format_byte_width (format);
-	  if (l > 0)
-	    whandle->dhandle.n_values = l;
-	  else
-	    success = FALSE;
-	}
-      else
-	success = FALSE;
-
-      if (!success)
-	gsl_data_handle_common_free (&whandle->dhandle);
-    }
-  if (!success)
-    {
-      gsl_delete_struct (WaveHandle, whandle);
-      return NULL;
-    }
-  gsl_data_handle_enter_cache (&whandle->dhandle, hash); /* FIXME: we never leave or handle mtime */
-
-  return &whandle->dhandle;
-}
-
-GslDataHandle*
-gsl_wave_handle_new (const gchar      *file_name,
-		     GTime             mtime, /* may be 0 */
-		     GslWaveFormatType format,
-		     guint             byte_order,
-		     GslLong           byte_offset,
-		     GslLong           n_values)
-{
-  GslDataHandle *whandle, *thandle;
-  guint fwidth, boffset_reminder;
-  GslLong offset;
-  
   g_return_val_if_fail (file_name != NULL, NULL);
   g_return_val_if_fail (format > GSL_WAVE_FORMAT_NONE && format < GSL_WAVE_FORMAT_LAST, NULL);
   g_return_val_if_fail (byte_order == G_LITTLE_ENDIAN || byte_order == G_BIG_ENDIAN, NULL);
   g_return_val_if_fail (byte_offset >= 0, NULL);
+  g_return_val_if_fail (n_channels >= 1, NULL);
   g_return_val_if_fail (n_values >= 1 || n_values == -1, NULL);
 
-  /* FIXME: handle mtime */
-
-  fwidth = wave_format_byte_width (format);
-  boffset_reminder = byte_offset % fwidth;
-  whandle = raw_wave_handle_new_cached (file_name, format, byte_order, boffset_reminder);
-  if (!whandle)
-    return NULL;
-
-  offset = (byte_offset - boffset_reminder) / fwidth;
-  if (n_values == -1)
-    n_values = whandle->n_values - offset;
-
-  if (offset + n_values > whandle->n_values)
+  whandle = gsl_new_struct0 (WaveHandle, 1);
+  if (gsl_data_handle_common_init (&whandle->dhandle, file_name))
     {
-      gsl_data_handle_unref (whandle);
+      whandle->dhandle.vtable = &wave_handle_vtable;
+      whandle->n_channels = n_channels;
+      whandle->format = format;
+      whandle->byte_order = byte_order;
+      whandle->byte_offset = byte_offset;
+      whandle->requested_length = n_values;
+      whandle->hfile = NULL;
+      return &whandle->dhandle;
+    }
+  else
+    {
+      gsl_delete_struct (WaveHandle, whandle);
       return NULL;
     }
-
-  /* do we need a translation handle? */
-  if (offset == 0 && whandle->n_values == n_values)
-    return whandle;
-
-  thandle = gsl_data_handle_new_translate (whandle,
-					   0, offset,
-					   whandle->n_values - offset - n_values);
-  gsl_data_handle_unref (whandle);
-  g_assert (thandle->n_values == n_values);	/* paranoid */
-  
-  return thandle;
 }
 
 const gchar*
@@ -1042,126 +1239,3 @@ gsl_wave_format_from_string (const gchar *string)
     return is_unsigned ? GSL_WAVE_FORMAT_UNSIGNED_16 : GSL_WAVE_FORMAT_SIGNED_16;
   return GSL_WAVE_FORMAT_NONE;
 }
-
-
-/* --- datahandle cache --- */
-typedef struct _DHCSlot DHCSlot;
-struct _DHCSlot
-{
-  DHCSlot *next;
-  GslDataHandle *dhandle;
-  GslDataHandleHash hash;
-};
-static DHCSlot *dhc_slots = NULL;
-static GslMutex dhc_mutex = { 0, };
-
-void
-gsl_data_handle_enter_cache (GslDataHandle    *dhandle,
-			     GslDataHandleHash hash)
-{
-  DHCSlot *slot;
-
-  g_return_if_fail (dhandle != NULL);
-  g_return_if_fail (dhandle->ref_count > 0);
-  g_return_if_fail (gsl_data_handle_cached (hash) == NULL);
-
-  slot = gsl_new_struct (DHCSlot, 1);
-  slot->hash = hash;
-  slot->hash.strings[0] = g_strdup (hash.strings[0]);
-  slot->hash.strings[1] = g_strdup (hash.strings[1]);
-  slot->hash.strings[2] = g_strdup (hash.strings[2]);
-  slot->hash.strings[3] = g_strdup (hash.strings[3]);
-  slot->dhandle = gsl_data_handle_ref (dhandle);
-
-  GSL_SPIN_LOCK (&dhc_mutex);
-  slot->next = dhc_slots;
-  dhc_slots = slot;
-  GSL_SPIN_UNLOCK (&dhc_mutex);
-}
-
-void
-gsl_data_handle_leave_cache (GslDataHandle *dhandle)
-{
-  DHCSlot *slot, *last;
-  
-  g_return_if_fail (dhandle != NULL);
-  g_return_if_fail (dhandle->ref_count > 0);
-
-  GSL_SPIN_LOCK (&dhc_mutex);
-  for (last = NULL, slot = dhc_slots; slot; last = slot, slot = last->next)
-    if (slot->dhandle == dhandle)
-      break;
-  if (slot)
-    {
-      if (last)
-	last->next = slot->next;
-      else
-	dhc_slots->next = slot->next;
-    }
-  GSL_SPIN_UNLOCK (&dhc_mutex);
-
-  if (slot)
-    {
-      g_free (slot->hash.strings[0]);
-      g_free (slot->hash.strings[1]);
-      g_free (slot->hash.strings[2]);
-      g_free (slot->hash.strings[3]);
-      gsl_data_handle_ref (slot->dhandle);
-      gsl_delete_struct (DHCSlot, slot);
-    }
-  else
-    g_warning (G_STRLOC ": data handle %p not in cache", dhandle);
-}
-
-static gboolean
-dhhash_match (GslDataHandleHash *hash1,
-	      GslDataHandleHash *hash2)
-{
-  if (hash1->fpointer != hash2->fpointer)
-    return FALSE;
-  if (memcmp (&hash1->data, &hash2->data, sizeof (hash1->data)) != 0)
-    return FALSE;
-  if ((hash1->strings[0] != NULL) ^ (hash2->strings[0] != NULL))
-    return FALSE;
-  if ((hash1->strings[1] != NULL) ^ (hash2->strings[1] != NULL))
-    return FALSE;
-  if ((hash1->strings[2] != NULL) ^ (hash2->strings[2] != NULL))
-    return FALSE;
-  if ((hash1->strings[3] != NULL) ^ (hash2->strings[3] != NULL))
-    return FALSE;
-  if (hash1->strings[0] && strcmp (hash1->strings[0], hash2->strings[0]) != 0)
-    return FALSE;
-  if (hash1->strings[1] && strcmp (hash1->strings[1], hash2->strings[1]) != 0)
-    return FALSE;
-  if (hash1->strings[2] && strcmp (hash1->strings[2], hash2->strings[2]) != 0)
-    return FALSE;
-  if (hash1->strings[3] && strcmp (hash1->strings[3], hash2->strings[3]) != 0)
-    return FALSE;
-  return TRUE;
-}
-
-GslDataHandle*
-gsl_data_handle_cached (GslDataHandleHash hash)
-{
-  DHCSlot *slot;
-
-  GSL_SPIN_LOCK (&dhc_mutex);
-  for (slot = dhc_slots; slot; slot = slot->next)
-    if (dhhash_match (&slot->hash, &hash))
-      break;
-  GSL_SPIN_UNLOCK (&dhc_mutex);
-  return slot ? gsl_data_handle_ref (slot->dhandle) : NULL;
-}
-
-
-/* --- initialization --- */
-void
-_gsl_init_data_handles (void)
-{
-  static gboolean initialized = FALSE;
-  
-  g_assert (initialized == FALSE);
-  initialized = TRUE;
-  
-  gsl_mutex_init (&dhc_mutex);
-}
