@@ -24,7 +24,7 @@
 #ifndef VIRTUALPORTS_H
 #define VIRTUALPORTS_H
 
-#include "synthschedule.h"
+#include "gslschedule.h"
 #include <list>
 
 /*
