@@ -166,4 +166,4 @@ SystemThreads::~SystemThreads()
 {
 }
 
-namespace Arts { void *gslGlobalMutexTable = 0; }
+namespace Arts { void ARTS_EXPORT *gslGlobalMutexTable = 0; }
