@@ -33,7 +33,7 @@ void adjust_priority()
 		
 		sp.sched_priority = priority;
 
-		if(sched_setscheduler(0, SCHED_FIFO, &sp) == 0)
+		if(sched_setscheduler(0, SCHED_FIFO, &sp) != -1)
 		{
 			printf(">> running as realtime process now (priority %ld)\n",
 																	priority);
@@ -95,6 +95,10 @@ int main(int argc, char **argv)
 #else
 		setreuid(-1, getuid());
 #endif
+		if (geteuid() != getuid()) {
+			perror("setuid()");
+			return 2;
+		}
 	}
 
 	if(argc == 0)
