@@ -81,7 +81,7 @@ public:
   		unsigned long nitems;
   		unsigned long bytes_after;
   		long offset;
-  		char *buf;
+  		unsigned char *buf;
 	
 		propertyMap.clear();
 
@@ -94,15 +94,15 @@ public:
   		{
 			XGetWindowProperty( X11display, rootWin, atom, offset, 256,
                         False, XA_STRING, &type, &format, &nitems, &bytes_after,
-                        (unsigned char **)&buf);
+                        &buf);
 
 			if(type == None)	// no property -> no contents
 				return;
 		
-    		s += buf;
+    		s += (char*) buf;
     		offset += nitems/4;
     		if (buf)
-      			XFree(buf);
+      			XFree((char*) buf);
   		}
 
   		// Parse through the property string stripping out key value pairs
